#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef USE_NLLIBC
#include <nllibc.h>
#endif

#include "lib.h"
#include "objlist.h"
#include "type.h"
#include "symbol.h"
#include "value.h"
#include "asm.h"
#include "asm_syntax.h"
#include "asm_symbol.h"

static int reg_next = 0;

static int reg_reserve(void)
{
  int r;
  r = reg_next++;
  return r;
}

static int reg_release(void)
{
  --reg_next;
  return reg_next;
}

static int reg_current(void)
{
  return reg_next - 1;
}

static char *reg_name(int reg)
{
  static char *name[] = {
    "Register Error",
    "%ebx",
    "64(%esp)",  "68(%esp)",  "72(%esp)",  "76(%esp)",
    "80(%esp)",  "84(%esp)",  "88(%esp)",  "92(%esp)",
    "96(%esp)",  "100(%esp)", "104(%esp)", "108(%esp)",
    "112(%esp)", "116(%esp)", "120(%esp)", "124(%esp)",
  };
  if (reg + 1 >= (sizeof(name) / sizeof(name[0])))
    ERREXIT(1);
  return name[reg + 1];
}

static char *reg_name_current(void)
{
  return reg_name(reg_current());
}

static int label_next = 1;

static int label_reserve(int n)
{
  int r;
  r = label_next;
  label_next += n;
  return r;
}

static int label_release(int n)
{
  label_next -= n;
  return label_next;
}

static int label_current(void)
{
  return label_next - 1;
}

static int reg_stack_size(void)
{
  return sizeof(int) * 32;
}

static char *r0_by_size(model_t model)
{
  switch (model_get_size(model)) {
  case -1: ERREXIT(1);
  case  1: return "al";
  case  2: return "ax";
  default: break;
  }
  return "eax";
}

static char *r1_by_size(model_t model)
{
  switch (model_get_size(model)) {
  case -1: ERREXIT(1);
  case  1: return "dl";
  case  2: return "dx";
  default: break;
  }
  return "edx";
}

static char *r2_by_size(model_t model)
{
  switch (model_get_size(model)) {
  case -1: ERREXIT(1);
  case  1: return "cl";
  case  2: return "cx";
  default: break;
  }
  return "ecx";
}

static char *mov_by_size(model_t model)
{
  return "mov";
}

static int sign_extension(FILE *out, model_t model)
{
  if (!model_is_address(model)) {
    switch (model->type) {
    case C_TYPE_CHAR:
      if (model->flags & MODEL_FLAG_UNSIGNED)
	fprintf(out, "\tmovzbl\t%%al,\t%%eax\n");
      else
	fprintf(out, "\tmovsbl\t%%al,\t%%eax\n");
      break;
    case C_TYPE_SHORT:
      if (model->flags & MODEL_FLAG_UNSIGNED)
	fprintf(out, "\tmovzwl\t%%ax,\t%%eax\n");
      else
	fprintf(out, "\tmovswl\t%%ax,\t%%eax\n");
      break;
    default:
      break;
    }
  }

  return 0;
}

int asm_symbol_address_get(FILE *out, symbol_t symbol)
{
  asm_section_change_text(out);

  fprintf(out, "\t/* ADDRESS GET BEGIN (%s, %s) */\n", symbol->name, type_get_word(symbol->type));

  saveline(symbol->line);

  switch (symbol->type) {
  case C_TYPE_LABEL:
    /* fall through */
  case C_TYPE_VARIABLE:
    switch (symbol->location.type) {
    case C_TYPE_EXTERN:
    case C_TYPE_STATIC:
      fprintf(out, "\tmov\t$%s, %%eax\n", symbol->name);
      break;
    case C_TYPE_STACK:
      fprintf(out, "\tlea\t%d(%%esp), %%eax\n",
	      reg_stack_size() + symbol->location.offset);
      break;
    case C_TYPE_STRUCT:
    case C_TYPE_UNION:
      ERREXIT(1);
    case C_TYPE_SYMBOL:
      asm_symbol_address_get(out, symbol->location.base.symbol);
      fprintf(out, "\tlea\t%d(%%eax), %%eax\n", symbol->location.offset);
      break;
    case C_TYPE_POINTER:
      asm_symbol_value_get(out, symbol->location.base.symbol);
      fprintf(out, "\tlea\t%d(%%eax), %%eax\n", symbol->location.offset);
      break;
    case C_TYPE_NONE:
      fprintf(out, "\tmov\t$%s, %%eax\n", symbol->name);
      break;
    default:
      ERREXIT(1);
    }
    break;
  default:
    ERREXIT(1);
  }

  fprintf(out, "\t/* ADDRESS GET END (%s) */\n", symbol->name);
  return 0;
}

static int asm_symbol_value_op(FILE *out, symbol_t symbol)
{
  symbol_t *arg;
  model_t model;
  int i, n;

  fprintf(out, "\t/* VALUE OP BEGIN (%s, %s, %s) */\n",
	  symbol->name, type_get_word(symbol->type),
	  type_get_word(symbol->value->type));

  switch (symbol->value->type) {
  case C_TYPE_NUMBER:
    fprintf(out, "\tmov\t$0x%x, %%eax\n", symbol->value->obj.number);
    goto ret;
  default:
    break;
  }

  arg = symbol->value->obj.op.arg;

  switch (symbol->value->type) {
  case C_TYPE_CAST:
    asm_symbol_value_set(out, arg[0]);
    asm_symbol_value_get(out, arg[0]);
    break;
  case C_TYPE_ADDRESS:
    asm_symbol_value_set(out, arg[0]);
    asm_symbol_address_get(out, arg[0]);
    break;
  case C_TYPE_POINTER:
    asm_symbol_value_set(out, arg[0]);
    asm_symbol_value_get(out, arg[0]);
    if (symbol->type != C_TYPE_LABEL) {
      fprintf(out, "\t%s\t(%%eax), %%%s\n", mov_by_size(symbol->model), r0_by_size(symbol->model));
    }
    break;
  case C_TYPE_DOT:
    asm_symbol_value_set(out, arg[0]);
    asm_symbol_address_get(out, arg[0]);
    fprintf(out, "\t%s\t%d(%%eax), %%%s\n",
	    mov_by_size(symbol->model), symbol->location.offset, r0_by_size(symbol->model));
    break;
  case C_TYPE_NOT:
    asm_symbol_value_set(out, arg[0]);
    asm_symbol_value_get(out, arg[0]);
    label_reserve(2);
    fprintf(out, "\tcmp\t$0, %%%s\n", r0_by_size(arg[0]->model));
    fprintf(out, "\tje\t%df\n", label_current() - 1);
    fprintf(out, "\tmov\t$0, %%eax\n");
    fprintf(out, "\tjmp\t%df\n", label_current());
    fprintf(out, "%d:\tmov\t$1, %%eax\n", label_current() - 1);
    fprintf(out, "%d:\n", label_current());
    label_release(2);
    break;
  case C_TYPE_INV:
    asm_symbol_value_set(out, arg[0]);
    asm_symbol_value_get(out, arg[0]);
    fprintf(out, "\tnot\t%%eax\n");
    break;
  case C_TYPE_PLUS:
    asm_symbol_value_set(out, arg[0]);
    asm_symbol_value_get(out, arg[0]);
    break;
  case C_TYPE_MINUS:
    asm_symbol_value_set(out, arg[0]);
    asm_symbol_value_get(out, arg[0]);
    fprintf(out, "\tneg\t%%eax\n");
    break;
  case C_TYPE_INC: /* ++x */
    asm_symbol_value_set(out, arg[0]);
    asm_symbol_address_get(out, arg[0]);
    model = arg[0]->model;
    fprintf(out, "\tmov\t%%eax, %%edx\n");
    fprintf(out, "\t%s\t(%%edx), %%%s\n", mov_by_size(model), r0_by_size(model));
    n = model_is_pointer(arg[0]->model) ? model_get_pointer_size(arg[0]->model) : 1;
    fprintf(out, "\tadd\t$%d, %%%s\n", n, r0_by_size(model));
    fprintf(out, "\t%s\t%%%s, (%%edx)\n", mov_by_size(model), r0_by_size(model));
    break;
  case C_TYPE_DEC: /* --x */
    asm_symbol_value_set(out, arg[0]);
    asm_symbol_address_get(out, arg[0]);
    model = arg[0]->model;
    fprintf(out, "\tmov\t%%eax, %%edx\n");
    fprintf(out, "\t%s\t(%%edx), %%%s\n", mov_by_size(model), r0_by_size(model));
    n = model_is_pointer(arg[0]->model) ? model_get_pointer_size(arg[0]->model) : 1;
    fprintf(out, "\tsub\t$%d, %%%s\n", n, r0_by_size(model));
    fprintf(out, "\t%s\t%%%s, (%%edx)\n", mov_by_size(model), r0_by_size(model));
    break;
  case C_TYPE_INCA: /* x++ */
    asm_symbol_value_set(out, arg[0]);
    asm_symbol_address_get(out, arg[0]);
    model = arg[0]->model;
    fprintf(out, "\tmov\t%%eax, %%edx\n");
    fprintf(out, "\t%s\t(%%edx), %%%s\n", mov_by_size(model), r0_by_size(model));
    reg_reserve();
    fprintf(out, "\tmov\t%%eax, %s\n", reg_name_current());
    n = model_is_pointer(arg[0]->model) ? model_get_pointer_size(arg[0]->model) : 1;
    fprintf(out, "\tadd\t$%d, %%%s\n", n, r0_by_size(model));
    fprintf(out, "\t%s\t%%%s, (%%edx)\n", mov_by_size(model), r0_by_size(model));
    fprintf(out, "\tmov\t%s, %%eax\n", reg_name_current());
    reg_release();
    break;
  case C_TYPE_DECA: /* x-- */
    asm_symbol_value_set(out, arg[0]);
    asm_symbol_address_get(out, arg[0]);
    model = arg[0]->model;
    fprintf(out, "\tmov\t%%eax, %%edx\n");
    fprintf(out, "\t%s\t(%%edx), %%%s\n", mov_by_size(model), r0_by_size(model));
    reg_reserve();
    fprintf(out, "\tmov\t%%eax, %s\n", reg_name_current());
    n = model_is_pointer(arg[0]->model) ? model_get_pointer_size(arg[0]->model) : 1;
    fprintf(out, "\tsub\t$%d, %%%s\n", n, r0_by_size(model));
    fprintf(out, "\t%s\t%%%s, (%%edx)\n", mov_by_size(model), r0_by_size(model));
    fprintf(out, "\tmov\t%s, %%eax\n", reg_name_current());
    reg_release();
    break;

  case C_TYPE_ADD:
    asm_symbol_value_set(out, arg[0]);
    asm_symbol_value_set(out, arg[1]);
    asm_symbol_value_get(out, arg[1]);
    model = arg[0]->model;
    reg_reserve();
    fprintf(out, "\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_value_get(out, arg[0]);
    fprintf(out, "\tmov\t%s, %%edx\n", reg_name_current());
    reg_release();
    fprintf(out, "\tadd\t%%%s, %%%s\n", r1_by_size(model), r0_by_size(model));
    break;
  case C_TYPE_SUB:
    asm_symbol_value_set(out, arg[0]);
    asm_symbol_value_set(out, arg[1]);
    asm_symbol_value_get(out, arg[1]);
    model = arg[0]->model;
    reg_reserve();
    fprintf(out, "\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_value_get(out, arg[0]);
    fprintf(out, "\tmov\t%s, %%edx\n", reg_name_current());
    reg_release();
    fprintf(out, "\tsub\t%%%s, %%%s\n", r1_by_size(model), r0_by_size(model));
    break;
  case C_TYPE_MUL:
    asm_symbol_value_set(out, arg[0]);
    asm_symbol_value_set(out, arg[1]);
    asm_symbol_value_get(out, arg[1]);
    model = arg[0]->model;
    reg_reserve();
    fprintf(out, "\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_value_get(out, arg[0]);
    fprintf(out, "\tmov\t%s, %%edx\n", reg_name_current());
    reg_release();
    fprintf(out, "\timul\t%%%s\n", r1_by_size(model));
    break;
  case C_TYPE_DIV:
    asm_symbol_value_set(out, arg[0]);
    asm_symbol_value_set(out, arg[1]);
    asm_symbol_value_get(out, arg[1]);
    model = arg[0]->model;
    reg_reserve();
    fprintf(out, "\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_value_get(out, arg[0]);
    fprintf(out, "\tmov\t%s, %%ecx\n", reg_name_current());
    reg_release();
    if (arg[0]->model->flags & MODEL_FLAG_UNSIGNED) {
      fprintf(out, "\tmov\t$0, %%edx\n");
    } else {
      fprintf(out, "\tcltd\n");
    }
    if (symbol->model->flags & MODEL_FLAG_UNSIGNED) {
      fprintf(out, "\tdiv\t%%%s\n", r2_by_size(arg[1]->model));
    } else {
      fprintf(out, "\tidiv\t%%ecx\n");
    }
    break;
  case C_TYPE_MOD:
    asm_symbol_value_set(out, arg[0]);
    asm_symbol_value_set(out, arg[1]);
    asm_symbol_value_get(out, arg[1]);
    model = arg[0]->model;
    reg_reserve();
    fprintf(out, "\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_value_get(out, arg[0]);
    fprintf(out, "\tmov\t%s, %%ecx\n", reg_name_current());
    reg_release();
    if (arg[0]->model->flags & MODEL_FLAG_UNSIGNED) {
      fprintf(out, "\tmov\t$0, %%edx\n");
    } else {
      fprintf(out, "\tcltd\n");
    }
    if (symbol->model->flags & MODEL_FLAG_UNSIGNED) {
      fprintf(out, "\tdiv\t%%ecx\n");
    } else {
      fprintf(out, "\tidiv\t%%ecx\n");
    }
    fprintf(out, "\tmov\t%%edx, %%eax\n");
    break;
  case C_TYPE_AND:
    asm_symbol_value_set(out, arg[0]);
    asm_symbol_value_set(out, arg[1]);
    asm_symbol_value_get(out, arg[1]);
    model = arg[0]->model;
    reg_reserve();
    fprintf(out, "\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_value_get(out, arg[0]);
    fprintf(out, "\tmov\t%s, %%edx\n", reg_name_current());
    reg_release();
    fprintf(out, "\tand\t%%%s, %%%s\n", r1_by_size(model), r0_by_size(model));
    break;
  case C_TYPE_OR:
    asm_symbol_value_set(out, arg[0]);
    asm_symbol_value_set(out, arg[1]);
    asm_symbol_value_get(out, arg[1]);
    model = arg[0]->model;
    reg_reserve();
    fprintf(out, "\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_value_get(out, arg[0]);
    fprintf(out, "\tmov\t%s, %%edx\n", reg_name_current());
    reg_release();
    fprintf(out, "\tor\t%%%s, %%%s\n", r1_by_size(model), r0_by_size(model));
    break;
  case C_TYPE_XOR:
    asm_symbol_value_set(out, arg[0]);
    asm_symbol_value_set(out, arg[1]);
    asm_symbol_value_get(out, arg[1]);
    model = arg[0]->model;
    reg_reserve();
    fprintf(out, "\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_value_get(out, arg[0]);
    fprintf(out, "\tmov\t%s, %%edx\n", reg_name_current());
    reg_release();
    fprintf(out, "\txor\t%%%s, %%%s\n", r1_by_size(model), r0_by_size(model));
    break;
  case C_TYPE_LSHIFT:
    asm_symbol_value_set(out, arg[0]);
    asm_symbol_value_set(out, arg[1]);
    asm_symbol_value_get(out, arg[1]);
    model = arg[0]->model;
    reg_reserve();
    fprintf(out, "\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_value_get(out, arg[0]);
    fprintf(out, "\tmov\t%s, %%ecx\n", reg_name_current());
    reg_release();
    fprintf(out, "\tshl\t%%cl, %%eax\n");
    break;
  case C_TYPE_RSHIFT:
    asm_symbol_value_set(out, arg[0]);
    asm_symbol_value_set(out, arg[1]);
    asm_symbol_value_get(out, arg[1]);
    model = arg[0]->model;
    reg_reserve();
    fprintf(out, "\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_value_get(out, arg[0]);
    fprintf(out, "\tmov\t%s, %%ecx\n", reg_name_current());
    reg_release();
    fprintf(out, "\t%s\t%%cl, %%eax\n",
	    (arg[0]->model->flags & MODEL_FLAG_UNSIGNED) ? "shr" : "sar");
    break;

  case C_TYPE_LT:
    asm_symbol_value_set(out, arg[0]);
    asm_symbol_value_set(out, arg[1]);
    asm_symbol_value_get(out, arg[1]);
    model = arg[0]->model;
    reg_reserve();
    fprintf(out, "\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_value_get(out, arg[0]);
    fprintf(out, "\tmov\t%s, %%edx\n", reg_name_current());
    reg_release();
    label_reserve(2);
    fprintf(out, "\tcmp\t%%%s, %%%s\n", r1_by_size(model), r0_by_size(model));
    if (model->flags & MODEL_FLAG_UNSIGNED) {
      fprintf(out, "\tjb\t%df\n", label_current() - 1);
    } else {
      fprintf(out, "\tjl\t%df\n", label_current() - 1);
    }
    fprintf(out, "\tmov\t$0, %%eax\n");
    fprintf(out, "\tjmp\t%df\n", label_current());
    fprintf(out, "%d:\tmov\t$1, %%eax\n", label_current() - 1);
    fprintf(out, "%d:\n", label_current());
    label_release(2);
    break;
  case C_TYPE_GT:
    asm_symbol_value_set(out, arg[0]);
    asm_symbol_value_set(out, arg[1]);
    asm_symbol_value_get(out, arg[1]);
    model = arg[0]->model;
    reg_reserve();
    fprintf(out, "\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_value_get(out, arg[0]);
    fprintf(out, "\tmov\t%s, %%edx\n", reg_name_current());
    reg_release();
    label_reserve(2);
    fprintf(out, "\tcmp\t%%%s, %%%s\n", r1_by_size(model), r0_by_size(model));
    if (model->flags & MODEL_FLAG_UNSIGNED) {
      fprintf(out, "\tja\t%df\n", label_current() - 1);
    } else {
      fprintf(out, "\tjg\t%df\n", label_current() - 1);
    }
    fprintf(out, "\tmov\t$0, %%eax\n");
    fprintf(out, "\tjmp\t%df\n", label_current());
    fprintf(out, "%d:\tmov\t$1, %%eax\n", label_current() - 1);
    fprintf(out, "%d:\n", label_current());
    label_release(2);
    break;
  case C_TYPE_LTEQ:
    asm_symbol_value_set(out, arg[0]);
    asm_symbol_value_set(out, arg[1]);
    asm_symbol_value_get(out, arg[1]);
    model = arg[0]->model;
    reg_reserve();
    fprintf(out, "\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_value_get(out, arg[0]);
    fprintf(out, "\tmov\t%s, %%edx\n", reg_name_current());
    reg_release();
    label_reserve(2);
    fprintf(out, "\tcmp\t%%%s, %%%s\n", r1_by_size(model), r0_by_size(model));
    if (model->flags & MODEL_FLAG_UNSIGNED) {
      fprintf(out, "\tjbe\t%df\n", label_current() - 1);
    } else {
      fprintf(out, "\tjle\t%df\n", label_current() - 1);
    }
    fprintf(out, "\tmov\t$0, %%eax\n");
    fprintf(out, "\tjmp\t%df\n", label_current());
    fprintf(out, "%d:\tmov\t$1, %%eax\n", label_current() - 1);
    fprintf(out, "%d:\n", label_current());
    label_release(2);
    break;
  case C_TYPE_GTEQ:
    asm_symbol_value_set(out, arg[0]);
    asm_symbol_value_set(out, arg[1]);
    asm_symbol_value_get(out, arg[1]);
    model = arg[0]->model;
    reg_reserve();
    fprintf(out, "\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_value_get(out, arg[0]);
    fprintf(out, "\tmov\t%s, %%edx\n", reg_name_current());
    reg_release();
    label_reserve(2);
    fprintf(out, "\tcmp\t%%%s, %%%s\n", r1_by_size(model), r0_by_size(model));
    if (model->flags & MODEL_FLAG_UNSIGNED) {
      fprintf(out, "\tjae\t%df\n", label_current() - 1);
    } else {
      fprintf(out, "\tjge\t%df\n", label_current() - 1);
    }
    fprintf(out, "\tmov\t$0, %%eax\n");
    fprintf(out, "\tjmp\t%df\n", label_current());
    fprintf(out, "%d:\tmov\t$1, %%eax\n", label_current() - 1);
    fprintf(out, "%d:\n", label_current());
    label_release(2);
    break;
  case C_TYPE_EQEQ:
    asm_symbol_value_set(out, arg[0]);
    asm_symbol_value_set(out, arg[1]);
    asm_symbol_value_get(out, arg[1]);
    model = arg[0]->model;
    reg_reserve();
    fprintf(out, "\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_value_get(out, arg[0]);
    fprintf(out, "\tmov\t%s, %%edx\n", reg_name_current());
    reg_release();
    label_reserve(2);
    fprintf(out, "\tcmp\t%%%s, %%%s\n", r1_by_size(model), r0_by_size(model));
    fprintf(out, "\tje\t%df\n", label_current() - 1);
    fprintf(out, "\tmov\t$0, %%eax\n");
    fprintf(out, "\tjmp\t%df\n", label_current());
    fprintf(out, "%d:\tmov\t$1, %%eax\n", label_current() - 1);
    fprintf(out, "%d:\n", label_current());
    label_release(2);
    break;
  case C_TYPE_NEQ:
    asm_symbol_value_set(out, arg[0]);
    asm_symbol_value_set(out, arg[1]);
    asm_symbol_value_get(out, arg[1]);
    model = arg[0]->model;
    reg_reserve();
    fprintf(out, "\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_value_get(out, arg[0]);
    fprintf(out, "\tmov\t%s, %%edx\n", reg_name_current());
    reg_release();
    label_reserve(2);
    fprintf(out, "\tcmp\t%%%s, %%%s\n", r1_by_size(model), r0_by_size(model));
    fprintf(out, "\tjne\t%df\n", label_current() - 1);
    fprintf(out, "\tmov\t$0, %%eax\n");
    fprintf(out, "\tjmp\t%df\n", label_current());
    fprintf(out, "%d:\tmov\t$1, %%eax\n", label_current() - 1);
    fprintf(out, "%d:\n", label_current());
    label_release(2);
    break;

  case C_TYPE_EQ:
    asm_symbol_value_set(out, arg[1]);
    asm_symbol_value_set(out, arg[0]);
    asm_symbol_value_get(out, arg[1]);
    model = arg[0]->model;
    reg_reserve();
    fprintf(out, "\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_address_get(out, arg[0]);
    fprintf(out, "\tmov\t%s, %%edx\n", reg_name_current());
    reg_release();
    fprintf(out, "\t%s\t%%%s, (%%eax)\n", mov_by_size(model), r1_by_size(model));
    fprintf(out, "\tmov\t%%edx, %%eax\n");
    break;
  case C_TYPE_ADDEQ:
    asm_symbol_value_set(out, arg[1]);
    asm_symbol_value_set(out, arg[0]);
    asm_symbol_value_get(out, arg[1]);
    model = arg[0]->model;
    reg_reserve();
    fprintf(out, "\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_value_get(out, arg[0]);
    fprintf(out, "\tmov\t%s, %%edx\n", reg_name_current());
    fprintf(out, "\tadd\t%%%s, %%%s\n", r1_by_size(model), r0_by_size(model));
    fprintf(out, "\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_address_get(out, arg[0]);
    fprintf(out, "\tmov\t%s, %%edx\n", reg_name_current());
    reg_release();
    fprintf(out, "\t%s\t%%%s, (%%eax)\n", mov_by_size(model), r1_by_size(model));
    fprintf(out, "\tmov\t%%edx, %%eax\n");
    break;
  case C_TYPE_SUBEQ:
    asm_symbol_value_set(out, arg[1]);
    asm_symbol_value_set(out, arg[0]);
    asm_symbol_value_get(out, arg[1]);
    model = arg[0]->model;
    reg_reserve();
    fprintf(out, "\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_value_get(out, arg[0]);
    fprintf(out, "\tmov\t%s, %%edx\n", reg_name_current());
    fprintf(out, "\tsub\t%%%s, %%%s\n", r1_by_size(model), r0_by_size(model));
    fprintf(out, "\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_address_get(out, arg[0]);
    fprintf(out, "\tmov\t%s, %%edx\n", reg_name_current());
    reg_release();
    fprintf(out, "\t%s\t%%%s, (%%eax)\n", mov_by_size(model), r1_by_size(model));
    fprintf(out, "\tmov\t%%edx, %%eax\n");
    break;
  case C_TYPE_MULEQ:
    asm_symbol_value_set(out, arg[1]);
    asm_symbol_value_set(out, arg[0]);
    asm_symbol_value_get(out, arg[1]);
    model = arg[0]->model;
    reg_reserve();
    fprintf(out, "\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_value_get(out, arg[0]);
    fprintf(out, "\tmov\t%s, %%edx\n", reg_name_current());
    fprintf(out, "\timul\t%%%s\n", r1_by_size(model));
    fprintf(out, "\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_address_get(out, arg[0]);
    fprintf(out, "\tmov\t%s, %%edx\n", reg_name_current());
    reg_release();
    fprintf(out, "\t%s\t%%%s, (%%eax)\n", mov_by_size(model), r1_by_size(model));
    fprintf(out, "\tmov\t%%edx, %%eax\n");
    break;
  case C_TYPE_DIVEQ:
    asm_symbol_value_set(out, arg[1]);
    asm_symbol_value_set(out, arg[0]);
    asm_symbol_value_get(out, arg[1]);
    model = arg[0]->model;
    reg_reserve();
    fprintf(out, "\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_value_get(out, arg[0]);
    fprintf(out, "\tmov\t%s, %%ecx\n", reg_name_current());
    if (arg[0]->model->flags & MODEL_FLAG_UNSIGNED) {
      fprintf(out, "\tmov\t$0, %%edx\n");
    } else {
      fprintf(out, "\tcltd\n");
    }
    if (symbol->model->flags & MODEL_FLAG_UNSIGNED) {
      fprintf(out, "\tdiv\t%%%s\n", r2_by_size(arg[1]->model));
    } else {
      fprintf(out, "\tidiv\t%%ecx\n");
    }
    sign_extension(out, symbol->model);
    fprintf(out, "\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_address_get(out, arg[0]);
    fprintf(out, "\tmov\t%s, %%edx\n", reg_name_current());
    reg_release();
    fprintf(out, "\t%s\t%%%s, (%%eax)\n", mov_by_size(model), r1_by_size(model));
    fprintf(out, "\tmov\t%%edx, %%eax\n");
    break;
  case C_TYPE_MODEQ:
    asm_symbol_value_set(out, arg[1]);
    asm_symbol_value_set(out, arg[0]);
    asm_symbol_value_get(out, arg[1]);
    model = arg[0]->model;
    reg_reserve();
    fprintf(out, "\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_value_get(out, arg[0]);
    fprintf(out, "\tmov\t%s, %%ecx\n", reg_name_current());
    if (arg[0]->model->flags & MODEL_FLAG_UNSIGNED) {
      fprintf(out, "\tmov\t$0, %%edx\n");
    } else {
      fprintf(out, "\tcltd\n");
    }
    if (symbol->model->flags & MODEL_FLAG_UNSIGNED) {
      fprintf(out, "\tdiv\t%%ecx\n");
    } else {
      fprintf(out, "\tidiv\t%%ecx\n");
    }
    fprintf(out, "\tmov\t%%edx, %%eax\n");
    sign_extension(out, symbol->model);
    fprintf(out, "\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_address_get(out, arg[0]);
    fprintf(out, "\tmov\t%s, %%edx\n", reg_name_current());
    reg_release();
    fprintf(out, "\t%s\t%%%s, (%%eax)\n", mov_by_size(model), r1_by_size(model));
    fprintf(out, "\tmov\t%%edx, %%eax\n");
    break;
  case C_TYPE_LSHIFTEQ:
    asm_symbol_value_set(out, arg[1]);
    asm_symbol_value_set(out, arg[0]);
    asm_symbol_value_get(out, arg[1]);
    model = arg[0]->model;
    reg_reserve();
    fprintf(out, "\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_value_get(out, arg[0]);
    fprintf(out, "\tmov\t%s, %%ecx\n", reg_name_current());
    fprintf(out, "\tshl\t%%cl, %%eax\n");
    fprintf(out, "\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_address_get(out, arg[0]);
    fprintf(out, "\tmov\t%s, %%edx\n", reg_name_current());
    reg_release();
    fprintf(out, "\t%s\t%%%s, (%%eax)\n", mov_by_size(model), r1_by_size(model));
    fprintf(out, "\tmov\t%%edx, %%eax\n");
    break;
  case C_TYPE_RSHIFTEQ:
    asm_symbol_value_set(out, arg[1]);
    asm_symbol_value_set(out, arg[0]);
    asm_symbol_value_get(out, arg[1]);
    model = arg[0]->model;
    reg_reserve();
    fprintf(out, "\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_value_get(out, arg[0]);
    fprintf(out, "\tmov\t%s, %%ecx\n", reg_name_current());
    fprintf(out, "\t%s\t%%cl, %%eax\n",
	   (arg[0]->model->flags & MODEL_FLAG_UNSIGNED) ? "shr" : "sar");
    fprintf(out, "\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_address_get(out, arg[0]);
    fprintf(out, "\tmov\t%s, %%edx\n", reg_name_current());
    reg_release();
    fprintf(out, "\t%s\t%%%s, (%%eax)\n", mov_by_size(model), r1_by_size(model));
    fprintf(out, "\tmov\t%%edx, %%eax\n");
    break;
  case C_TYPE_ANDEQ:
    asm_symbol_value_set(out, arg[1]);
    asm_symbol_value_set(out, arg[0]);
    asm_symbol_value_get(out, arg[1]);
    model = arg[0]->model;
    reg_reserve();
    fprintf(out, "\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_value_get(out, arg[0]);
    fprintf(out, "\tmov\t%s, %%edx\n", reg_name_current());
    fprintf(out, "\tand\t%%%s, %%%s\n", r1_by_size(model), r0_by_size(model));
    fprintf(out, "\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_address_get(out, arg[0]);
    fprintf(out, "\tmov\t%s, %%edx\n", reg_name_current());
    reg_release();
    fprintf(out, "\t%s\t%%%s, (%%eax)\n", mov_by_size(model), r1_by_size(model));
    fprintf(out, "\tmov\t%%edx, %%eax\n");
    break;
  case C_TYPE_OREQ:
    asm_symbol_value_set(out, arg[1]);
    asm_symbol_value_set(out, arg[0]);
    asm_symbol_value_get(out, arg[1]);
    model = arg[0]->model;
    reg_reserve();
    fprintf(out, "\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_value_get(out, arg[0]);
    fprintf(out, "\tmov\t%s, %%edx\n", reg_name_current());
    fprintf(out, "\tor\t%%%s, %%%s\n", r1_by_size(model), r0_by_size(model));
    fprintf(out, "\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_address_get(out, arg[0]);
    fprintf(out, "\tmov\t%s, %%edx\n", reg_name_current());
    reg_release();
    fprintf(out, "\t%s\t%%%s, (%%eax)\n", mov_by_size(model), r1_by_size(model));
    fprintf(out, "\tmov\t%%edx, %%eax\n");
    break;
  case C_TYPE_XOREQ:
    asm_symbol_value_set(out, arg[1]);
    asm_symbol_value_set(out, arg[0]);
    asm_symbol_value_get(out, arg[1]);
    model = arg[0]->model;
    reg_reserve();
    fprintf(out, "\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_value_get(out, arg[0]);
    fprintf(out, "\tmov\t%s, %%edx\n", reg_name_current());
    fprintf(out, "\txor\t%%%s, %%%s\n", r1_by_size(model), r0_by_size(model));
    fprintf(out, "\tmov\t%%eax, %s\n", reg_name_current());
    asm_symbol_address_get(out, arg[0]);
    fprintf(out, "\tmov\t%s, %%edx\n", reg_name_current());
    reg_release();
    fprintf(out, "\t%s\t%%%s, (%%eax)\n", mov_by_size(model), r1_by_size(model));
    fprintf(out, "\tmov\t%%edx, %%eax\n");
    break;

  case C_TYPE_ANDAND:
    asm_symbol_value_set(out, arg[0]);
    asm_symbol_value_get(out, arg[0]);
    label_reserve(2);
    fprintf(out, "\tcmp\t$0, %%%s\n", r0_by_size(arg[0]->model));
    fprintf(out, "\tje\t%df\n", label_current() - 1);
    asm_symbol_value_set(out, arg[1]);
    asm_symbol_value_get(out, arg[1]);
    fprintf(out, "\tcmp\t$0, %%%s\n", r0_by_size(arg[1]->model));
    fprintf(out, "\tje\t%df\n", label_current() - 1);
    fprintf(out, "\tmov\t$1, %%eax\n");
    fprintf(out, "\tjmp\t%df\n", label_current());
    fprintf(out, "%d:\tmov\t$0, %%eax\n", label_current() - 1);
    fprintf(out, "%d:\n", label_current());
    label_release(2);
    break;
  case C_TYPE_OROR:
    asm_symbol_value_set(out, arg[0]);
    asm_symbol_value_get(out, arg[0]);
    label_reserve(2);
    fprintf(out, "\tcmp\t$0, %%%s\n", r0_by_size(arg[0]->model));
    fprintf(out, "\tjne\t%df\n", label_current() - 1);
    asm_symbol_value_set(out, arg[1]);
    asm_symbol_value_get(out, arg[1]);
    fprintf(out, "\tcmp\t$0, %%%s\n", r0_by_size(arg[1]->model));
    fprintf(out, "\tjne\t%df\n", label_current() - 1);
    fprintf(out, "\tmov\t$0, %%eax\n");
    fprintf(out, "\tjmp\t%df\n", label_current());
    fprintf(out, "%d:\tmov\t$1, %%eax\n", label_current() - 1);
    fprintf(out, "%d:\n", label_current());
    label_release(2);
    break;

  case C_TYPE_QUESTION:
    asm_symbol_value_set(out, arg[0]);
    asm_symbol_value_get(out, arg[0]);
    label_reserve(2);
    fprintf(out, "\tcmp\t$0, %%%s\n", r0_by_size(arg[0]->model));
    fprintf(out, "\tje\t%df\n", label_current() - 1);
    asm_symbol_value_set(out, arg[1]);
    asm_symbol_value_get(out, arg[1]);
    fprintf(out, "\tjmp\t%df\n", label_current());
    fprintf(out, "%d:\n", label_current() - 1);
    asm_symbol_value_set(out, arg[2]);
    asm_symbol_value_get(out, arg[2]);
    fprintf(out, "%d:\n", label_current());
    label_release(2);
    break;
  case C_TYPE_COMMA:
    asm_symbol_value_set(out, arg[0]);
    asm_symbol_value_set(out, arg[1]);
    asm_symbol_value_get(out, arg[0]);
    asm_symbol_value_get(out, arg[1]);
    break;

  case C_TYPE_ARG:
    asm_symbol_value_get(out, symbol);
    break;
  case C_TYPE_FUNCTION:
    if (model_is_pointer(symbol->value->obj.function.symbol->model)) {
      asm_symbol_value_set(out, symbol->value->obj.function.symbol);
      asm_symbol_value_get(out, symbol->value->obj.function.symbol);
      reg_reserve();
      fprintf(out, "\tmov\t%%eax, %s\n", reg_name_current());
    }
    for (i = 0; i < symbol->value->obj.function.args.number; i++) {
      asm_symbol_value_set(out, symbol->value->obj.function.args.arg[i]);
      asm_symbol_value_get(out, symbol->value->obj.function.args.arg[i]);
      reg_reserve();
      fprintf(out, "\tmov\t%%eax, %s\n", reg_name_current());
    }
    for (i = 0; i < symbol->value->obj.function.args.number; i++) {
      fprintf(out, "\tmov\t%s, %%eax\n", reg_name_current());
      fprintf(out, "\tmov\t%%eax, 0x%x(%%esp)\n",
	      (symbol->value->obj.function.args.number - 1 - i) * 4);
      reg_release();
    }

    if (model_is_pointer(symbol->value->obj.function.symbol->model)) {
      fprintf(out, "\tmov\t%s, %%eax\n", reg_name_current());
      reg_release();
      fprintf(out, "\tcall\t*%%eax\t/* %s */\n", symbol->value->obj.function.symbol->name);
    } else {
      fprintf(out, "\tcall\t%s\n", symbol->value->obj.function.symbol->name);
    }
    break;
  default:
    ERREXIT(1);
  }

ret:
  sign_extension(out, symbol->model);
  fprintf(out, "\t/* VALUE OP END (%s, %s, %s) */\n",
	  symbol->name, type_get_word(symbol->type),
	  type_get_word(symbol->value->type));
  return 0;
}

int asm_symbol_value_set(FILE *out, symbol_t symbol)
{
  asm_section_change_text(out);

  fprintf(out, "\t/* VALUE SET BEGIN (%s, %s) */\n",
	  symbol->name, type_get_word(symbol->type));

  saveline(symbol->line);

  switch (symbol->type) {
  case C_TYPE_LABEL:
    if (symbol->flags & SYMBOL_FLAG_TEMPORARY) {
      asm_symbol_value_op(out, symbol);
    }
    break;

  case C_TYPE_VARIABLE:
    if (symbol->flags & SYMBOL_FLAG_TEMPORARY) {
      asm_symbol_value_op(out, symbol);
      if ((symbol->flags & SYMBOL_FLAG_LOCATED) &&
	  (symbol->location.type == C_TYPE_STACK)) {
	reg_reserve();
	fprintf(out, "\tmov\t%%eax, %s\n", reg_name_current());
	asm_symbol_address_get(out, symbol);
	fprintf(out, "\tmov\t%s, %%edx\n", reg_name_current());
	reg_release();
	fprintf(out, "\t%s\t%%%s, (%%eax)\n",
		mov_by_size(symbol->model), r1_by_size(symbol->model));
      }
    }
    break;
  default:
    ERREXIT(1);
  }

  fprintf(out, "\t/* VALUE SET END (%s) */\n", symbol->name);
  return 0;
}

int asm_symbol_value_get(FILE *out, symbol_t symbol)
{
  asm_section_change_text(out);

  fprintf(out, "\t/* VALUE GET BEGIN (%s, %s) */\n",
	  symbol->name, type_get_word(symbol->type));

  saveline(symbol->line);

  switch (symbol->type) {
  case C_TYPE_LABEL:
    asm_symbol_address_get(out, symbol);
    break;
  case C_TYPE_VARIABLE:
    if (symbol->flags & SYMBOL_FLAG_READONLY) {
      switch (symbol->value->type) {
      case C_TYPE_NUMBER:
	fprintf(out, "\t%s\t$%d, %%%s\n", mov_by_size(symbol->model),
		symbol->value->obj.number, r0_by_size(symbol->model));
	break;
      default:
	ERREXIT(1);
      }
      break;
    }
    if ((symbol->flags & SYMBOL_FLAG_LOCATED) ||
	(symbol->scope == C_TYPE_EXTERN)) {
      asm_symbol_address_get(out, symbol);
      fprintf(out, "\t%s\t(%%eax), %%%s\n",
	      mov_by_size(symbol->model), r0_by_size(symbol->model));
      break;
    }
    ERREXIT(1);
  default:
    ERREXIT(1);
  }

  sign_extension(out, symbol->model);
  fprintf(out, "\t/* VALUE GET END (%s) */\n", symbol->name);
  return 0;
}

static int asm_symbol_label_value(FILE *out, symbol_t symbol);
static int asm_symbol_variable_value(FILE *out, symbol_t symbol);

static int asm_symbol_array(FILE *out, objlist_t symbols)
{
  objentry_t entry;
  symbol_t symbol;

  for (entry = objlist_get_head(symbols);
       !objlist_is_end(symbols, entry);
       entry = objentry_get_next(entry)) {
    symbol = objentry_get_obj(entry);
#if 1
    fprintf(out, "\t.align\t%d\n", model_get_align(symbol->model));
    switch (symbol->type) {
    case C_TYPE_LABEL:    asm_symbol_label_value(out, symbol);    break;
    case C_TYPE_VARIABLE: asm_symbol_variable_value(out, symbol); break;
    default:
      ERREXIT(1);
    }
#else
    asm_symbol(out, symbol);
#endif
  }

  return 0;
}

static int asm_symbol_area(FILE *out, area_t area)
{
  fprintf(out, "\t.align\t%d\n", area->align);
  fprintf(out, "\t.space\t%d, 0\n", area->size);
  return 0;
}

static int asm_symbol_function(FILE *out, symbol_t symbol)
{
  syntax_t syntax;
  objentry_t entry;
  symbol_t s;
  int n;

  asm_section_change_data(out);

  fprintf(out, "%s:\n", symbol->obj.function.initlabel->obj.label.name);
  asm_symbol_array(out, symbol->obj.function.stack->symbols);
  fprintf(out, "\n");

  asm_section_change_text(out);

  saveline(symbol->line);

  if (symbol->scope != C_TYPE_STATIC)
    fprintf(out, "\t.globl\t%s\n", symbol->name);
  fprintf(out, "\t.type\t%s, @function\n", symbol->name);
  fprintf(out, "%s:\n", symbol->name);

  if ((symbol->value != NULL) &&
      (symbol->value->type == C_TYPE_BLOCK)) {
    syntax = symbol->value->obj.block.syntax;
    if (syntax->type == C_TYPE_BLOCK) {
      fprintf(out, "\tpush\t%%ebp\n");
      fprintf(out, "\tmov\t%%esp, %%ebp\n");
      fprintf(out, "\tpush\t%%esi\n");
      fprintf(out, "\tpush\t%%edi\n");
      fprintf(out, "\tpush\t%%ebx\n");

      fprintf(out, "\tsub\t$0x%x, %%esp\n",
	      reg_stack_size() + symbol->obj.function.stack->size);

      fprintf(out, "\tlea\t0x%x(%%esp), %%ebx\n", reg_stack_size());
      fprintf(out, "\tmov\t$%s, %%ecx\n", symbol->obj.function.initlabel->obj.label.name);
      fprintf(out, "\tmov\t$0x%x, %%edx\n", symbol->obj.function.stack->size);
      label_reserve(2);
      fprintf(out, "%d:", label_current() - 1);
      fprintf(out, "\tcmp\t$0, %%edx\n");
      fprintf(out, "\tjle\t%df\n", label_current());
      fprintf(out, "\tmov\t(%%ecx), %%al\n");
      fprintf(out, "\tinc\t%%ecx\n");
      fprintf(out, "\tmov\t%%al, (%%ebx)\n");
      fprintf(out, "\tinc\t%%ebx\n");
      fprintf(out, "\tdec\t%%edx\n");
      fprintf(out, "\tjmp\t%db\n", label_current() - 1);
      fprintf(out, "%d:", label_current());
      label_release(2);

      n = 0;
      for (entry = objlist_get_head(symbol->obj.function.args);
	   !objlist_is_end(symbol->obj.function.args, entry);
	   entry = objentry_get_next(entry)) {
	s = objentry_get_obj(entry);
	fprintf(out, "\tmov\t0x%x(%%esp), %%eax\n",
		reg_stack_size() + symbol->obj.function.stack->size
		+ 4 /* for EBX */
		+ 8 /* for EDI/ESI */
		+ 4 /* for EBP */
		+ 4 /* for return address */
		+ (n++) * 4);
	fprintf(out, "\tmov\t%%eax, 0x%x(%%esp)\n",
		reg_stack_size() + s->location.offset);
      }
      asm_syntax_list(out, syntax->obj.block.syntaxes);
      fprintf(out, "%s:\n", symbol->obj.function.retlabel->obj.label.name);
      fprintf(out, "\tadd\t$0x%x, %%esp\n",
	      reg_stack_size() + symbol->obj.function.stack->size);
      fprintf(out, "\tpop\t%%ebx\n");
      fprintf(out, "\tpop\t%%edi\n");
      fprintf(out, "\tpop\t%%esi\n");
      fprintf(out, "\tleave\n");
      fprintf(out, "\tret\n\n");
    }
  }

  return 0;
}

static int asm_symbol_label_header(FILE *out, symbol_t symbol)
{
  asm_section_change_data(out);

  saveline(symbol->line);

  if (symbol->scope != C_TYPE_STATIC)
    fprintf(out, "\t.globl\t%s\n", symbol->name);
  fprintf(out, "\t.align\t%d\n", model_get_align(symbol->model));
  fprintf(out, "\t.type\t%s, @object\n", symbol->name);
  fprintf(out, "\t.size\t%s, %d\n", symbol->name, model_get_size(symbol->model));

  fprintf(out, "%s:\n", symbol->name);

  return 0;
}

static int asm_symbol_string(FILE *out, value_t value)
{
  int i;

  fprintf(out, "\t.string\t\"");
  for (i = 0; i < value->obj.string.size - 1; i++) {
    switch (value->obj.string.s[i]) {
    case '\0': fprintf(out, "\\0"); break;
    case '\t': fprintf(out, "\\t"); break;
    case '\n': fprintf(out, "\\n"); break;
    case '\v': fprintf(out, "\\v"); break;
    case '\f': fprintf(out, "\\f"); break;
    case '\r': fprintf(out, "\\r"); break;
    case '\'': fprintf(out, "\\\'"); break;
    case '\"': fprintf(out, "\\\""); break;
    case '\\': fprintf(out, "\\\\"); break;
    default:   fprintf(out, "%c", value->obj.string.s[i]); break;
    }
  }
  fprintf(out, "\"\n");

  return 0;
}

static int asm_symbol_label_value(FILE *out, symbol_t symbol)
{
  if (symbol->value == NULL) {
    fprintf(out, "\t.space\t%d, 0\t/* %s */\n", model_get_size(symbol->model), symbol->name);
  } else {
    switch (symbol->value->type) {
    case C_TYPE_STRING:
      asm_symbol_string(out, symbol->value);
      break;
    case C_TYPE_ARRAY:
      asm_symbol_array(out, symbol->value->obj.array.values);
      fprintf(out, "\n");
      break;
    default:
      ERREXIT(1);
    }
  }

  return 0;
}

static int asm_symbol_label(FILE *out, symbol_t symbol)
{
  asm_symbol_label_header(out, symbol);
  asm_symbol_label_value(out, symbol);
  return 0;
}

static int asm_symbol_variable_header(FILE *out, symbol_t symbol)
{
  asm_section_change_data(out);

  saveline(symbol->line);

  if (symbol->scope != C_TYPE_STATIC)
    fprintf(out, "\t.globl\t%s\n", symbol->name);
  fprintf(out, "\t.align\t%d\n", model_get_align(symbol->model));
  fprintf(out, "\t.type\t%s, @object\n", symbol->name);
  fprintf(out, "\t.size\t%s, %d\n", symbol->name, model_get_size(symbol->model));

  fprintf(out, "%s:\n", symbol->name);

  return 0;
}

static int asm_symbol_variable_value(FILE *out, symbol_t symbol)
{
  int size;

  if (model_is_pointer(symbol->model)) {
    fprintf(out, "\t.long");
  } else {
    switch (symbol->model->type) {
    case C_TYPE_CHAR:       fprintf(out, "\t.byte" ); break;
    case C_TYPE_SHORT:      fprintf(out, "\t.short"); break;
    case C_TYPE_INT:        fprintf(out, "\t.int"  ); break;
    case C_TYPE_ENUM:       fprintf(out, "\t.int"  ); break;
    case C_TYPE_LONG:       fprintf(out, "\t.long" ); break;
    case C_TYPE_LONGLONG:   fprintf(out, "\t.long" ); break;
    case C_TYPE_FLOAT:      fprintf(out, "\t.long" ); break;
    case C_TYPE_DOUBLE:     fprintf(out, "\t.long" ); break;
    case C_TYPE_LONGDOUBLE: fprintf(out, "\t.long" ); break;
    case C_TYPE_STRUCT:
      if (symbol->value != NULL) {
	fprintf(out, "\n");
	asm_symbol_array(out, symbol->value->obj.array.values);
      } else {
	asm_symbol_area(out, symbol->model->custom._struct->members);
      }
      return 0;
    case C_TYPE_UNION:
      asm_symbol_area(out, symbol->model->custom._union->members);
      return 0;
    default:
      ERREXIT(1);
    }
  }

  fprintf(out, "\t");

  if ((symbol->value == NULL) || (symbol->flags & SYMBOL_FLAG_TEMPORARY)) {
    fprintf(out, "0");
  } else {
    switch (symbol->value->type) {
    case C_TYPE_ARG:     fprintf(out, "0"); break;
    case C_TYPE_NUMBER:  fprintf(out, "%d", symbol->value->obj.number); break;
    case C_TYPE_LABEL:   fprintf(out, "%s", symbol->value->obj.label.symbol->name); break;
    case C_TYPE_ADDRESS: fprintf(out, "%s", symbol->value->obj.label.symbol->name); break;
    default:
      ERREXIT(1);
    }
  }

  if (symbol->name != NULL)
    fprintf(out, "\t/* %s */", symbol->name);
  fprintf(out, "\n");

  if (!model_is_pointer(symbol->model)) {
    switch (symbol->model->type) {
    case C_TYPE_LONGLONG:
    case C_TYPE_FLOAT:
    case C_TYPE_DOUBLE:
    case C_TYPE_LONGDOUBLE:
      size = model_get_size(symbol->model) - 4;
      for (; size > 0; size -= 4)
	fprintf(out, "\t.long\t0\t/* dummy */\n");
      break;
    default:
      break;
    }
  }

  return 0;
}

static int asm_symbol_variable(FILE *out, symbol_t symbol)
{
  asm_symbol_variable_header(out, symbol);
  asm_symbol_variable_value(out, symbol);
  fprintf(out, "\n");
  return 0;
}

int asm_symbol(FILE *out, symbol_t symbol)
{
  saveline(symbol->line);

  switch (symbol->type) {
  case C_TYPE_LABEL:
    if (symbol->flags & SYMBOL_FLAG_FUNCTION) {
      if (symbol->flags & SYMBOL_FLAG_FUNCBLOCK)
	asm_symbol_function(out, symbol);
      break;
    }
    asm_symbol_label(out, symbol);
    break;
  case C_TYPE_VARIABLE:
    asm_symbol_variable(out, symbol);
    break;
  default:
    ERREXIT(1);
  }

  return 0;
}
